#!/usr/bin/env python3
# coding=utf-8

# Python implementation of stream library
# https://github.com/vgteam/stream
#
# The MIT License (MIT)
#
# Copyright (c) 2016 Ali Ghaffaari
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Test stream module."""

import os
import gzip
import filecmp
from context import stream
import vg_pb2


def read_alns1(fpath):
    """Read Alignment objects from a (GAM) file by using `with` statement.

    Args:
        fpath (string): path of the file to be read.
    """
    alns_list = []
    with stream.open(fpath, "rb") as istream:
        for data in istream:
            aln = vg_pb2.Alignment()
            aln.ParseFromString(data)
            alns_list.append(aln)
    return alns_list


def read_alns2(fpath):
    """Read Alignment objects from a (GAM) file without using `with` statement.

    Args:
        fpath (string): path of the file to be read.
    """
    alns_list = []
    istream = stream.open(fpath, "rb")
    for data in istream:
        aln = vg_pb2.Alignment()
        aln.ParseFromString(data)
        alns_list.append(aln)
    istream.close()
    return alns_list


def write_alns1(fpath, *objs_list):
    """Write 'Alignment' objects into the file by using `with` statement. It
    writes half of them in one group, and then the other half in another group.

    Args:
        fpath (string): path of the file to be written.
        objs_list (tuple: protobuf objects): list of objects to be written.
    """
    with stream.open(fpath, "wb") as ostream:
        length = len(objs_list)
        ostream.write(*objs_list[:length//2])
        ostream.write(*objs_list[length//2:])


def write_alns2(fpath, *objs_list):
    """Write 'Alignment' objects into the file without using `with` statement.
    It writes half of them in one group, and the other half in another group.

    Args:
        fpath (string): path of the file to be written.
        objs_list (tuple: protobuf objects): list of objects to be written.
    """
    ostream = stream.open(fpath, "wb", buffer_size=6)
    length = len(objs_list)
    ostream.write(*objs_list[:length//2])
    ostream.write(*objs_list[length//2:])
    ostream.close()


def test_all():
    """Run all test cases."""
    # Files
    testdir = os.path.dirname(os.path.realpath(__file__))
    gamfile = os.path.join(testdir, "sample_reads.gam")
    gamfile_nof_alns = 12
    rw1_gamfile = os.path.join(testdir, "rw1_sample_reads.gam")
    rw2_gamfile = os.path.join(testdir, "rw2_sample_reads.gam")
    # GUM file == Unzipped GAM file
    rw1_gumfile = os.path.join(testdir, "rw1_sample_reads.gum")
    rw2_gumfile = os.path.join(testdir, "rw2_sample_reads.gum")

    # Read a sample file.
    alns = read_alns1(gamfile)
    assert len(alns) == gamfile_nof_alns
    # Rewrite it into a new file in two groups of 6 objects.
    write_alns1(rw1_gamfile, *alns)
    # Read the rewritted file.
    re_alns = read_alns2(rw1_gamfile)
    # Check the length of the objects storing in both files.
    assert len(alns) == len(re_alns)
    # Rewrite again the read data.
    write_alns2(rw2_gamfile, *re_alns)
    # Unzip two generated files.
    with gzip.open(rw1_gamfile, "rb") as gfp, \
            open(rw1_gumfile, "wb") as ufp:
        ufp.write(gfp.read())
    with gzip.open(rw2_gamfile, "rb") as gfp, \
            open(rw2_gumfile, "wb") as ufp:
        ufp.write(gfp.read())
    # Check whether the two generated files have the same the content.
    assert filecmp.cmp(rw1_gumfile, rw2_gumfile)
    # Delete the generated files.
    os.remove(rw1_gamfile)
    os.remove(rw1_gumfile)
    os.remove(rw2_gamfile)
    os.remove(rw2_gumfile)

if __name__ == "__main__":
    test_all()



# coding=utf-8

"""Include release information of the 'pystream-protobuf' package."""

# CONSTANTS ####################################################################
# Development statuses:
DS_PLANNING = 1
DS_PREALPHA = 2
DS_ALPHA = 3
DS_BETA = 4
DS_STABLE = 5
DS_MATURE = 6
DS_INACTIVE = 7
DS_STRING = {
    DS_PLANNING: "Development Status :: 1 - Planning",
    DS_PREALPHA: "Development Status :: 2 - Pre-Alpha",
    DS_ALPHA: "Development Status :: 3 - Alpha",
    DS_BETA: "Development Status :: 4 - Beta",
    DS_STABLE: "Development Status :: 5 - Production/Stable",
    DS_MATURE: "Development Status :: 6 - Mature",
    DS_INACTIVE: "Development Status :: 7 - Inactive"
}
################################################################################

# Package release information.
__title__ = "stream"
__description__ = "Python implementation of stream library"
__author__ = "Ali Ghaffaari"
__email__ = "ali.ghaffaari@mpi-inf.mpg.de"
__license__ = "MIT"

# Release
__version__ = "1.2.0rc0"
__status__ = DS_ALPHA

# Sub-packages
__subpackages__ = []

# PyPI-related information
__keywords__ = ['protobuf', 'stream', 'protocol buffer']
__classifiers__ = [
    # Development status
    DS_STRING[__status__],

    # License
    'License :: OSI Approved :: MIT License',

    # Supported Python versions.
    'Programming Language :: Python :: 3.3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',

    # Intended Audience and Topic
    'Intended Audience :: Developers',
]
__requires__ = ['protobuf==3.0.0']
__tests_require__ = [
    'nose==1.3.7',
]



# coding=utf-8

# Python implementation of stream library
# https://github.com/vgteam/stream
#
# The MIT License (MIT)
#
# Copyright (c) 2016 Ali Ghaffaari
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Python implementation of stream library (https://github.com/vgteam/stream)
parsing all files encoded by stream and writing protobuf message instances into
the file by the same encoding.
"""

import gzip
from google.protobuf.internal.decoder import _DecodeVarint as varintDecoder
from google.protobuf.internal.encoder import _EncodeVarint as varintEncoder


def open(fpath, mode='rb', buffer_size=-1):  # pylint: disable=redefined-builtin
    """Open an stream."""
    return Stream(fpath, mode, buffer_size)


class Stream(object):
    """Stream class.

    Read and write protocol buffer streams encoded by 'stream' library. Stream
    objects instantiated for reading by setting mode to 'rb' (input `Stream`s)
    are iterable. So, protobuf objects can be obtained by iterate over the
    Stream. Stream iterator yields protobuf encoded data, so it should be parsed
    by using proper methods in Google Protocol Buffer library (for example
    `ParseFromString()` method).

    In output `Stream`s (those are instantiated with 'w' mode), method `write()`
    groups the given list of protobuf objects and writes them into the stream in
    the same format which is readable by any other parsers (refer to the stream
    library documentation for further information).

    The stream should be closed after performing all stream operations. Streams
    can be also used by `with` statement just like files.

    Here's a sample code using the Stream class to read from a file (so-called
    GAM file) containing a set of VG's (https://github.com/vgteam/vg) Alignment
    objects (defined: https://github.com/vgteam/vg/blob/master/src/vg.proto). It
    yields the protobuf objects stored in the file:

        import stream
        import vg_pb2

        alns_list = []
        with stream.open("test.gam", "rb") as istream:
            for data in istream:
                aln = vg_pb2.Alignment()
                aln.ParseFromString(data)
                alns_list.append(aln)

    Or

        import stream
        import vg_pb2

        alns_list = []
        istream = stream.open("test.gam", "rb")
        for data in istream:
            aln = vg_pb2.Alignment()
            aln.ParseFromString(data)
            alns_list.append(aln)
        istream.close()

    And here is another sample code for writing multiple protobuf objects into a
    file (here a GAM file):

        with stream.open("test.gam", "wb") as ostream:
            ostream.write(*objects_list)
            ostream.write(*another_objects_list)

    Or

        ostream = stream.open("test.gam", "wb")
        ostream.write(*objects_list)
        ostream.write(*another_objects_list)
        ostream.close()
    """
    def __init__(self, fpath, mode='rb', buffer_size=-1):
        """Constructor for the Stream class.

        Args:
            fpath (string): Path of the working file.
            mode (string): The mode argument can be any of 'r', 'rb', 'a', 'ab',
                'w', or 'wb', depending on whether the file will be read or
                written. The default is 'rb'.
            buffer_size (int): Write buffer size. The objects will be buffered
                before writing. No buffering will be made if buffer_size is -1.
        """
        self._fd = gzip.open(fpath, mode)
        self._buffer_size = buffer_size
        self._write_buff = []

    def __enter__(self):
        """Enter the runtime context related to Stream class. It will be
        automatically run by `with` statement.
        """
        return self

    def __exit__(self, *args):
        """Exit the runtime context related to Stream class. It will be
        automatically run by `with` statement. It closes the stream.
        """
        self.close()

    def __iter__(self):
        """Return the iterator object of the stream."""
        return self._get_objs()

    def _read_varint(self):
        """Read a varint from file, parse it, and return the decoded integer."""
        buff = self._fd.read(1)
        if buff == b'':
            return 0

        while (buff[-1] & 0x80) >> 7 == 1:  # while the MSB is 1
            new_byte = self._fd.read(1)
            if new_byte == b'':
                raise EOFError('unexpected EOF.')
            buff += new_byte

        varint, _ = varintDecoder(buff, 0)

        return varint

    def _get_objs(self):
        """A generator yielding all protobuf object data in the file. It is the
        main parser of the stream encoding.
        """
        while True:
            count = self._read_varint()
            if count == 0:
                break
            # Read a group containing `count` number of objects.
            for _ in range(count):
                size = self._read_varint()
                if size == 0:
                    raise EOFError('unexpected EOF.')
                # Read an object from the object group.
                yield self._fd.read(size)

    def close(self):
        """Close the stream."""
        self.flush()
        self._fd.close()

    def write(self, *pb2_obj):
        """Write a group of one or more protobuf objects to the file. Multiple
        object groups can be written by calling this method several times before
        closing stream or exiting the runtime context.

        The input protobuf objects get buffered and will be written down when
        the number of buffered objects exceed the `self._buffer_size`.
        """
        count = len(self._write_buff)
        if count >= self._buffer_size:
            self.flush()

        for obj in pb2_obj:
            self._write_buff.append(obj)

    def flush(self):
        """Write down buffer to the file."""
        count = len(self._write_buff)
        if count == 0:
            return

        varintEncoder(self._fd.write, count)

        for obj in self._write_buff:
            obj_str = obj.SerializeToString()
            varintEncoder(self._fd.write, len(obj_str))
            self._fd.write(obj_str)

        self._write_buff = []



# coding=utf-8

"""Setup script."""

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

import stream

_PYPINAME = "pystream-protobuf"
_PACKAGES = [stream.release.__title__] + stream.release.__subpackages__
_GITHUB_BASE = "https://github.com/cartoonist/"
_VCS_URL = _GITHUB_BASE + _PYPINAME
_TAR_URL = _VCS_URL + "/tarball/" + stream.release.__version__


setup(
    name=_PYPINAME,
    packages=_PACKAGES,
    version=stream.release.__version__,
    description=stream.release.__description__,
    author=stream.release.__author__,
    author_email=stream.release.__email__,
    license=stream.release.__license__,
    url=_VCS_URL,
    download_url=_TAR_URL,
    keywords=stream.release.__keywords__,
    classifiers=stream.release.__classifiers__,
    install_requires=stream.release.__requires__,
    tests_require=stream.release.__tests_require__,
)



# coding=utf-8

"""Generate a 'requirements.txt' file based on the release information defined
in `stream.release` module. It particularly reads `install_requires` and
`tests_require` fields.
"""

import os


def get_release_globals():
    """Read release module and extract `__requires__` and `__tests_require__`.

    This script is executed when required packages are not installed. So, the
    package cannot be imported. This function reads the release.py in stream
    package and extract `__requires__` and `__tests_require__` global variables
    from the script.

    Return:
        tuple(__requires__, __tests_require__)
    """
    with open(os.path.join('stream', 'release.py')) as release_file:
        exec(release_file.read(), globals())         # pylint: disable=exec-used
        return (__requires__,          # noqa pylint: disable=undefined-variable
                __tests_require__)     # noqa pylint: disable=undefined-variable


def write_reqfile():
    """Read release information and generate requirements file."""
    with open('requirements.txt', 'wt') as reqfile:
        requires, tests_require = get_release_globals()
        # Write install requirements.
        for req in requires:
            reqfile.write(req)
            reqfile.write('\n')

        # Write tests requirements.
        for treq in tests_require:
            reqfile.write(treq)
            reqfile.write('\n')


if __name__ == '__main__':
    write_reqfile()
